import asyncclick as click

import idac_sdk.cli.commands.status as status_set
import idac_sdk.cli.commands.restart as restart_set
import idac_sdk.cli.commands.cleanup as cleanup_set
import idac_sdk.cli.commands.create as create_set
import idac_sdk.cli.commands.extend as extend_set
import idac_sdk.cli.commands.config as config_set
from idac_sdk._version import __version__ as sdk_version


@click.group()
async def cli():
    """Group for CLI commands"""
    pass


@cli.command(short_help="show version")
async def version():
    click.echo(f"iDAC SDK v{sdk_version}")


def main():
    m = click.CommandCollection(
        sources=[
            cli,
            create_set.commands,
            cleanup_set.commands,
            status_set.commands,
            restart_set.commands,
            extend_set.commands,
            config_set.commands,
        ]
    )
    m(_anyio_backend="asyncio")


if __name__ == "__main__":
    main()



from enum import Enum, unique


@unique
class IDACRequestType(Enum):
    """iDAC request types"""

    SIMPLE = 0
    STATELESS = 1
    STATEFUL = 2


@unique
class IDACRequestStatus(Enum):
    """iDAC request statuses"""

    active = "active"
    queued = "queued"
    toBeDeleted = "toBeDeleted"
    update = "update"
    extend = "extend"
    scheduled = "scheduled"
    starting = "starting"
    cleaning = "cleaning"
    error = "error"
    cancelled = "cancelled"
    onboardError = "onboardError"
    complete = "complete"
    executed = "executed"
    deleted = "deleted"


@unique
class IDACAuthType(Enum):
    """iDAC authentication types"""

    NONE = 0
    BASIC = 1
    BEARER = 2
    DCLOUD_SESSION = 3
    WORKER = 4



from typing import Any, Generic, List, Optional, TypeVar, Union, Type, cast
from functools import wraps
from httpx import URL

from idac_sdk.base.controller import BaseIDACController
from idac_sdk.asynced.controller import IDACController as AsyncController
from idac_sdk.synced.controller import IDACController as SyncController
from idac_sdk.models.vpn_config import VPNConfig, VPNType
from idac_sdk.types import IDACAuthType, IDACRequestStatus
from idac_sdk.log import logger
from idac_sdk.errors import IncorrectWantedStateError, NoControllerError, UnknownVPNType
from idac_sdk.session_data import SessionData
from . import DEFAULT_USER_AGENT


def check_controller(method):

    @wraps(method)
    def wrapper(self, *method_args, **method_kwargs):
        if not isinstance(self.controller, BaseIDACController):
            raise NoControllerError("iDAC controller not provided")
        if (hasattr(self, "wanted_controller_type") and self.wanted_controller_type and
                not isinstance(self.controller, self.wanted_controller_type)):
            raise NoControllerError("Incorrect iDAC controller provided")
        return method(self, *method_args, **method_kwargs)

    return wrapper


T = TypeVar("T", AsyncController, SyncController)


class BaseIDACRequest(Generic[T]):
    session_data: Optional[SessionData]
    controller: T
    uuid: Optional[str]
    user_agent: str
    vpn: VPNConfig
    wanted_controller_type: Optional[Type[T]]
    max_redirects = 10

    def __init__(
        self,
        uuid: Optional[str] = None,
        session_data: Optional[SessionData] = None,
        controller: Optional[T] = None,
        user_agent: Optional[str] = None,
        vpn: Optional[VPNConfig] = None,
        wanted_controller_type: Optional[Type[T]] = None,
    ) -> None:
        """
        Base IDACRequest Object

        IDACRequest describes iDAC request object.
        Handles all operations with requests: create, cleanup, restart.

        Args:
            uuid (Optional[str], optional): Request UUID. Should be provided to work with existing
                requests Defaults to None.
            session_data (Optional[SessionData], optional): SessionData object. Will be send to iDAC
                controller Defaults to None.
            controller (Optional[IDACController], optional): IDACCOntroller object.
                Defaults to None.
            user_agent (Optional[str], optional): User-Agent string. Defaults to None.
            vpn (Optional[VPNConfig], optional): VPN options
        """
        if not uuid:
            # no UUID -> new request, need to set session_data and controller
            self.uuid = None
        else:
            # UUID provided -> existing request
            self.uuid = uuid

        self.session_data = SessionData() if not session_data else session_data

        self.user_agent = user_agent if user_agent else DEFAULT_USER_AGENT
        self.wanted_controller_type = wanted_controller_type

        if not controller and controller is not False:
            if not self.wanted_controller_type:
                raise NoControllerError("iDAC controller not provided")
            if isinstance(self.session_data, SessionData) and self.session_data.has("creds"):
                self.controller = self.wanted_controller_type(
                    auth_type=IDACAuthType.DCLOUD_SESSION,
                    auth=self.session_data.get("creds"),
                )
            else:
                self.controller = self.wanted_controller_type()
        else:
            self.controller = controller

        self.vpn = vpn or self.controller.vpn

    def add_vpn(self, where: dict[str, Any]) -> dict[str, Any]:
        """Adds VPN parameters to request

        Args:
            where (dict[str, Any]): Where VPN params should be added

        Raises:
            UnknownVPNType: Raised if unknown VPN type configured

        Returns:
            dict[str, Any]: updated dictionary with VPN parameters
        """
        logger.debug("Adding VPN parameters: %s", self.vpn)
        if ("idacVpn" in where) or (self.vpn.type == VPNType.none):
            return where

        if self.vpn.type == VPNType.vpod:
            # adding params for vPod VPN. `datacenter`, `vpod` and `anycpwd` should be loaded from session.xml
            where.update({"idacVpn": "dcloud-vpod"})
        elif self.vpn.type == VPNType.request or self.vpn.type == VPNType.explicit:
            # adding params for `request` - explicit VPN settings
            where.update({
                "idacVpn": "request",
                "idacVpnHost": self.vpn.params.host,
                "idacVpnUsername": self.vpn.params.username,
                "idacVpnPassword": self.vpn.params.password,
            })
        elif self.vpn.type == VPNType.secure_repo:
            # adding params for `secure_repo` - VPN settings will be loaded from secure repo
            where.update({
                "idacVpn": "secureRepo",
                "idacVpnSecureRepoSpace": self.vpn.params.secure_repo_space,
                "idacVpnSecureRepoFile": self.vpn.params.secure_repo_file,
                "idacVpnSecureRepoBlock": self.vpn.params.secure_repo_block,
                "idacVpnSecureRepoKey": self.vpn.params.secure_repo_key,
            })
        else:
            raise UnknownVPNType("Unknown VPN type provided")

        return where

    def sanitize_states(
        self, states: Union[List[IDACRequestStatus], List[str], List[Union[IDACRequestStatus, str]]]
    ) -> List[IDACRequestStatus]:
        for idx, val in enumerate(states):
            if isinstance(val, str):
                cast(List[IDACRequestStatus], states)[idx] = IDACRequestStatus(val)
            elif not isinstance(val, IDACRequestStatus):
                raise IncorrectWantedStateError(f"Incorrect wanted state: {val}")

        return cast(List[IDACRequestStatus], states)

    def raise_if_error(self, url: URL | str) -> None:
        err_detect = "notifications/error.php"
        if ((isinstance(url, str) and err_detect in url) or
            (isinstance(url, URL) and err_detect in url.path)):
            u = url if isinstance(url, URL) else URL(url)
            p = u.params
            msg: str = p.get("msg", None)
            if not msg:
                raise Exception(f"Unknown error during request execution for URL {u}")

            parts = msg.split(';')
            raise Exception("; ".join(parts))



import asyncclick as click

from idac_sdk import IDACRequestAsync
from idac_sdk.cli.lib._controller_options import _controller_options, with_controller
from idac_sdk.cli.lib._helpers import (
    add_options,
    _debug_options,
    set_logger_level,
)


@click.group()
def commands():
    pass


@commands.command(short_help="cleanup a request")
@add_options(_controller_options)
@add_options(_debug_options)
@click.option("-f", "--force", is_flag=True, default=False, help="Use force cleanup")
@click.argument("request_id")
@with_controller
async def cleanup(controller, request_id, force, **kwargs):
    """Execute STOP tasks of a request

    REQUEST_ID - id/uuid of the request"""
    set_logger_level(kwargs.get("debug", False))

    req = IDACRequestAsync(uuid=request_id, controller=controller)
    if force:
        await req.force_cleanup()
        click.echo("Started force cleanup")
    else:
        await req.cleanup()
        click.echo("Started cleanup")



from base64 import b64encode
from functools import cache
import hashlib
import secrets
from time import time
import json


def get_worker_hash(client_id: str, secret: str) -> str:
    timestamp = time.time()
    nonce = secrets.token_urlsafe()
    token_payload = {
        'client_id': client_id,
        'timestamp': timestamp,
        'access_token': hash_token(f"{secret}.{timestamp}.{nonce}"),
        'nonce': nonce
    }
    return json.dumps(token_payload)


@cache
def hash_token(payload: str) -> str:
    hasher = hashlib.sha3_256()
    hasher.update(payload.encode())
    digest = b64encode(hasher.digest()).decode()
    return digest



import asyncclick as click
from functools import wraps
from os import path

from idac_sdk.cli.lib._helpers import filter_dict
from idac_sdk import VPNConfig, VPNType

_vpn_options = [
    click.option(
        "--vpn",
        type=click.Choice([e.name for e in VPNType], case_sensitive=False),
        default=None,
        help="Enables VPN",
    ),
    # for explicit
    click.option(
        "--vpn-host", type=str, default=None, help="VPN host (for explicit/request VPN type)"
    ),
    click.option(
        "--vpn-username",
        type=str,
        default=None,
        help="VPN username (for explicit/request VPN type)",
    ),
    click.option(
        "--vpn-password",
        type=str,
        default=None,
        help="VPN password (for explicit/request VPN type)",
    ),
    # for secure repo
    click.option("--vpn-file", type=str, default=None, help="VPN file (for secure_repo VPN type)"),
    click.option(
        "--vpn-block", type=str, default=None, help="Block in VPN file (for secure_repo VPN type)"
    ),
    click.option(
        "--vpn-key", type=str, default=None, help="Key in the VPN block (for secure_repo VPN type)"
    ),
]


def with_vpn(func):
    @wraps(func)
    def wrapper(
        *args, vpn, vpn_host, vpn_username, vpn_password, vpn_file, vpn_block, vpn_key, **kwargs
    ):
        if vpn_file:
            secure_repo_space = f"{path.dirname(vpn_file)}/"
            secure_repo_file = path.basename(vpn_file)
        else:
            secure_repo_space = None
            secure_repo_file = None

        vpn_config = (
            None
            if not vpn
            else VPNConfig(
                **filter_dict(
                    {
                        "type": vpn or VPNType.none.name,
                        "params": filter_dict(
                            {
                                "host": vpn_host,
                                "username": vpn_username,
                                "password": vpn_password,
                                "secure_repo_space": secure_repo_space,
                                "secure_repo_file": secure_repo_file,
                                "secure_repo_block": vpn_block,
                                "secure_repo_key": vpn_key,
                            },
                            lambda elem: bool(elem[1]),
                        ),
                    },
                    lambda elem: bool(elem[1]),
                )
            )
        )
        return func(*args, vpn=vpn_config, **kwargs)

    return wrapper



import asyncclick as click

from idac_sdk import IDACRequestAsync
from idac_sdk.cli.lib._controller_options import _controller_options, with_controller
from idac_sdk.cli.lib._helpers import (
    add_options,
    _debug_options,
    set_logger_level,
)


@click.group()
def commands():
    pass


@commands.command(short_help="extend a request")
@add_options(_controller_options)
@add_options(_debug_options)
@click.option("-m", "--minutes", type=click.INT, help="Amount of minutes", required=True)
@click.argument("request_id")
@with_controller
async def extend(controller, request_id, minutes, **kwargs):
    """Extend request

    \b
    REQUEST_ID - id/uuid of the request"""
    set_logger_level(kwargs.get("debug", False))

    req = IDACRequestAsync(uuid=request_id, controller=controller)
    await req.extend(minutes)
    click.echo("Extended")



from base64 import b64encode
from email.policy import default
import yaml
import asyncclick as click
import idac_sdk
import idac_sdk.config
from idac_sdk.cli.lib._controller_options import _controller_options
from idac_sdk.cli.lib._vpn_options import _vpn_options
from idac_sdk.cli.lib._helpers import add_options
from idac_sdk.models.vpn_config import VPNType


@click.group()
def commands():
    pass


@commands.command(short_help="configure iDAC SDK")
@add_options(_controller_options)
@click.option(
    "--vpn",
    type=click.Choice([VPNType.none.name, VPNType.vpod.name], case_sensitive=False),
    default=None,
    help="Sets default VPN type",
)
async def config(
    controller_url,
    controller_proto,
    auth_type,
    auth,
    worker_id,
    worker_secret,
    api_version,
    vpn,
):
    current_config = idac_sdk.config.load_config(raise_error=False)
    if not controller_url:
        controller_url = click.prompt(
            "Enter default iDAC URL",
            type=str,
            default=current_config.defaults.idac_fqdn,
        )

    if not controller_proto:
        controller_proto = click.prompt(
            "Enter default iDAC controller proto",
            type=click.Choice(["http", "https"], case_sensitive=False),
            show_choices=True,
            default=current_config.defaults.idac_proto,
        )

    if not api_version:
        api_version = click.prompt(
            "Enter default iDAC controller API version",
            type=str,
            default=current_config.defaults.api_version,
        )

    if not auth_type:
        auth_type = click.prompt(
            "Enter default iDAC auth type",
            type=click.Choice([e.name for e in idac_sdk.IDACAuthType], case_sensitive=False),
            show_choices=True,
            default=current_config.defaults.auth.type,
        )
    if auth and auth_type not in [
        idac_sdk.IDACAuthType.DCLOUD_SESSION.name,
        idac_sdk.IDACAuthType.NONE.name,
    ]:
        params = {"token": auth}
    elif auth_type == idac_sdk.IDACAuthType.BASIC.name:
        username = click.prompt(
            "Enter username for BASIC auth",
            type=str,
            hide_input=False,
            default="",
            value_proc=lambda val: val if val else "",
        )
        password = click.prompt(
            "Enter password for BASIC auth",
            type=str,
            hide_input=True,
            default="",
            value_proc=lambda val: val if val else "",
        )
        params = {"token": b64encode(bytes(f"{username}:{password}", "utf-8")).decode("utf-8")}
    elif auth_type == idac_sdk.IDACAuthType.BEARER.name:
        if not auth:
            auth = click.prompt(
                "Enter default API token for BEARER auth",
                type=str,
                hide_input=True,
                default="",
                value_proc=lambda val: val if val else "",
            )
        params = {"token": auth}
    elif auth_type == idac_sdk.IDACAuthType.WORKER.name:
        params = {}
        if not worker_id:
            worker_id = click.prompt(
                "Enter client id for WORKER auth",
                type=str,
                hide_input=False,
                default="",
                value_proc=lambda val: val if val else "",
            )
        params["client_id"] = worker_id
        if not worker_secret:
            worker_secret = click.prompt(
                "Enter secret for WORKER auth",
                type=str,
                hide_input=True,
                default="",
                value_proc=lambda val: val if val else "",
            )
        params["secret"] = worker_secret
    else:
        params = {}
    auth = {"type": auth_type, "params": params}

    # VPN settings
    if not vpn:
        vpn = click.prompt(
            "Default VPN type (Only 'none' and 'vpod' can be stored in config)",
            type=click.Choice([VPNType.none.name, VPNType.vpod.name], case_sensitive=False),
            default=current_config.defaults.vpn.type.name,
        )
    vpn_config = {"type": vpn, "params": {}}

    with open(idac_sdk.config.IDAC_CONFIG_FILE, "w") as file:
        file.write(
            yaml.dump(
                {
                    "version": idac_sdk.config.IDAC_CONFIG_VERSION,
                    "defaults": {
                        "idac_fqdn": controller_url,
                        "idac_proto": controller_proto,
                        "auth": auth,
                        "api_version": api_version,
                        "vpn": vpn_config,
                    },
                }
            )
        )

    click.echo("Done")



import asyncclick as click
from functools import wraps

from idac_sdk.cli.lib._helpers import filter_dict
from idac_sdk import IDACAuthType, IDACControllerAsync

_controller_options = [
    click.option("-U", "--controller-url", default=None, type=str, help="URL of iDAC controller"),
    click.option(
        "-P",
        "--controller-proto",
        type=click.Choice(["http", "https"], case_sensitive=False),
        default="https",
        help="Which protocol should be used to connect to controller",
    ),
    click.option(
        "-T",
        "--auth-type",
        type=click.Choice([e.name for e in IDACAuthType], case_sensitive=False),
        help="Type of authentication",
    ),
    click.option(
        "-A",
        "--auth",
        type=str,
        help=(
            "Token if Auth Type is BEARER or WORKER, base64 string if BASIC. Ignored for other auth"
            " types"
        ),
    ),
    click.option(
        "--worker-id",
        type=str,
        help="Worker id for the Worker auth type",
    ),
    click.option(
        "--worker-secret",
        type=str,
        help="Worker secret for the Worker auth type",
    ),
    click.option("-V", "--api-version", default=None, help="API version to use"),
]


def with_controller(func):
    @wraps(func)
    def wrapper(
        *args,
        controller_url,
        controller_proto,
        auth_type,
        auth,
        worker_id,
        worker_secret,
        api_version,
        **kwargs
    ):
        controller = IDACControllerAsync(
            **filter_dict(
                {
                    "proto": controller_proto,
                    "url": controller_url,
                    "auth_type": IDACAuthType[auth_type] if auth_type else None,
                    "auth": auth,
                    "worker_id": worker_id,
                    "worker_secret": worker_secret,
                    "api_version": api_version,
                },
                lambda elem: bool(elem[1]),
            )
        )
        return func(*args, controller=controller, **kwargs)

    return wrapper



from httpx import AsyncClient, Request, ReadTimeout
import json
import asyncio
from typing import Dict, List, Optional, Tuple, Union

from idac_sdk.base import (
    DEFAULT_WAIT_INTERVAL,
    DEFAULT_WAIT_TIMEOUT,
    REQUEST_ERROR_STATES,
    REQUEST_GOOD_STATES,
)
from idac_sdk.base.request import BaseIDACRequest, check_controller
from idac_sdk.types import IDACRequestStatus, IDACRequestType, IDACAuthType
from idac_sdk.log import logger
from idac_sdk.asynced.controller import IDACController as AsyncController
from idac_sdk.errors import (
    IDACRequestStateError,
    IncorrectMinutesValue,
    NoAuth,
    NoIdError,
)
from idac_sdk.models.vpn_config import VPNConfig
from idac_sdk.models.request_state import RequestState
from idac_sdk.session_data import SessionData


class IDACRequest(BaseIDACRequest[AsyncController]):

    def __init__(
        self,
        uuid: Optional[str] = None,
        session_data: Optional[SessionData] = None,
        controller: Optional[AsyncController] = None,
        user_agent: Optional[str] = None,
        vpn: Optional[VPNConfig] = None,
    ) -> None:
        """
        Async IDACRequest Object

        IDACRequest describes iDAC request object.
        Handles all operations with requests: create, cleanup, restart.

        Args:
            uuid (Optional[str], optional): Request UUID. Should be provided to work with existing
                requests Defaults to None.
            session_data (Optional[SessionData], optional): SessionData object. Will be send to iDAC
                controller Defaults to None.
            controller (Optional[IDACController], optional): IDACCOntroller object.
                Defaults to None.
            user_agent (Optional[str], optional): User-Agent string. Defaults to None.
            vpn (Optional[VPNConfig], optional): VPN options
        """
        super().__init__(
            uuid=uuid,
            session_data=session_data,
            controller=controller,
            user_agent=user_agent,
            vpn=vpn,
            wanted_controller_type=AsyncController,
        )

    @check_controller
    async def check_controller_auth(self) -> None:
        """Tells controller to request auth token if auth type is DCLOUD_SESSION"""
        logger.debug(f"Checking auth. Type is {self.controller.auth.type}")
        if self.controller.auth.type == IDACAuthType.DCLOUD_SESSION.name:
            assert self.session_data, "Session data not set"
            if not self.session_data.has("creds"):
                raise NoAuth("No credentials set for DCLOUD_SESSION authentication.")
            token = await self.controller.get_auth_token(self.session_data.get("creds"),
                                                         self.session_data.get("datacenter"))
            logger.debug(f"dCloud token is: {token}")
        elif self.controller.auth.type == IDACAuthType.WORKER.name:
            token = await self.controller.get_auth_token()
            logger.debug(f"Worker token is: {token}")

    @check_controller
    async def create(
        self,
        request_type: IDACRequestType = IDACRequestType.SIMPLE
    ) -> Tuple[RequestState, Optional[str]]:
        """Sends Create API request. All data taken from session_data

        Args:
            request_type (IDACRequestType, optional): Type of the request.
                Defaults to IDACRequestType.SIMPLE.

        Returns:
            RequestState: State of the new request
        """
        assert self.session_data, "Session data not set"
        await self.check_controller_auth()

        if request_type == IDACRequestType.STATEFUL:
            api, method = self.controller.api_create_stateful()
        elif request_type == IDACRequestType.STATELESS:
            api, method = self.controller.api_create_stateless()
        else:
            api, method = self.controller.api_create()

        headers: Dict[str, str] = {"Accept": "application/json"}
        self.controller.with_auth(headers)
        logger.debug(f"API: {method} to {api}")
        logger.debug(f"Headers are: {headers}")

        # convert SessionData to dict and populate VPN parameters
        data = self.add_vpn(self.session_data.dict(exclude_none=True))

        if method == "POST":
            # put data in body if POST
            kwargs = {"json": data}
            headers.update({"Content-Type": "application/json"})
        else:
            # put data in query string if GET
            kwargs = {"params": data}

        redirect: str | None = None
        st: RequestState | None = None
        async with AsyncClient() as client:
            client.headers.update({"User-Agent": self.user_agent})

            # initial request
            request = Request(method, api, headers=headers, **kwargs)

            redirects = 0
            while redirects < self.max_redirects + 1:
                if redirects == self.max_redirects:
                    raise Exception(f"Exceeded max number of redirects - {self.max_redirects}")
                r = await client.send(request=request)
                logger.debug(f"Got response: {r.status_code} {r.reason_phrase}")
                if r.status_code < 200 or r.status_code > 399:
                    r.raise_for_status()

                json_body: dict
                try:
                    json_body = json.loads(r.text)
                except Exception as e:
                    logger.debug(f"Failed to parse JSON body: {e}")
                    if r.next_request:
                        self.raise_if_error(r.next_request.url)
                        logger.debug(f"Have next request, going there: {r.next_request.url}")
                        # update request and increase redirects counter
                        request = r.next_request
                        redirects += 1
                        continue
                    else:
                        raise e

                st = RequestState(**json_body)
                if st.request:
                    # grab UUID of a new automation
                    self.uuid = st.request.uuid

                if r.status_code >= 300 and r.status_code < 400:
                    # redirected
                    redirect = r.headers.get("location")
                break
            if not st:
                raise Exception("Failed to get request state")

        return st, redirect

    @check_controller
    async def get_state(self) -> RequestState:
        """Loads state of a request from controller

        Raises:
            NoIdError: if no ID/UUID set

        Returns:
            RequestState: State of the request
        """
        if not self.uuid:
            raise NoIdError("No ID/UUID provided for Get State")
        await self.check_controller_auth()

        api, method = self.controller.api_get_state(self.uuid)
        logger.debug(f"API: {method} to {api}")

        headers: Dict[str, str] = {"Accept": "application/json"}
        self.controller.with_auth(headers)

        async with AsyncClient() as client:
            client.headers.update({"User-Agent": self.user_agent})
            r = await client.request(method, api, headers=headers)
            if r.status_code < 200 or r.status_code > 399:
                r.raise_for_status()

            json_body: dict = json.loads(r.text)
            return RequestState(**json_body)

    @check_controller
    async def restart(self) -> None:
        """Restarts request

        Raises:
            NoIdError: if no ID/UUID set
        """
        if not self.uuid:
            raise NoIdError("No ID/UUID provided for Restart")
        await self.check_controller_auth()

        api, method = self.controller.api_restart(self.uuid)
        logger.debug(f"API: {method} to {api}")

        headers: Dict[str, str] = {"Accept": "application/json"}
        self.controller.with_auth(headers)

        async with AsyncClient() as client:
            client.headers.update({"User-Agent": self.user_agent})
            r = await client.request(method, api, headers=headers)
            logger.debug(f"Got response: {r.status_code} {r.reason_phrase}")
            if r.status_code < 200 or r.status_code > 399:
                r.raise_for_status()
            return

    @check_controller
    async def cleanup(self) -> None:
        """Cleans request

        Raises:
            NoIdError: if no ID/UUID set
        """
        if not self.uuid:
            raise NoIdError("No ID/UUID provided for Cleanup")
        await self.check_controller_auth()

        api, method = self.controller.api_cleanup(self.uuid)
        logger.debug(f"API: {method} to {api}")

        headers: Dict[str, str] = {"Accept": "application/json"}
        self.controller.with_auth(headers)

        async with AsyncClient() as client:
            client.headers.update({"User-Agent": self.user_agent})
            r = await client.request(method, api, headers=headers)
            logger.debug(f"Got response: {r.status_code} {r.reason_phrase}")
            if r.status_code < 200 or r.status_code > 399:
                r.raise_for_status()
            return

    @check_controller
    async def force_cleanup(self) -> None:
        """Forcibly cleans request

        Raises:
            NoIdError: if no ID/UUID set
        """
        if not self.uuid:
            raise NoIdError("No ID/UUID provided for Force Cleanup")
        await self.check_controller_auth()

        api, method = self.controller.api_force_cleanup(self.uuid)
        logger.debug(f"API: {method} to {api}")

        headers: Dict[str, str] = {"Accept": "application/json"}
        self.controller.with_auth(headers)

        async with AsyncClient() as client:
            client.headers.update({"User-Agent": self.user_agent})
            r = await client.request(method, api, headers=headers)
            logger.debug(f"Got response: {r.status_code} {r.reason_phrase}")
            if r.status_code < 200 or r.status_code > 399:
                r.raise_for_status()
            return

    @check_controller
    async def extend(self, minutes: int) -> None:
        """Extend a request by `minutes` minutes

        Args:
            minutes (int): amount of minutes

        Raises:
            NoIdError: if no ID provided
            IncorrectMinutesValue: if incorrect amount of minutes provided
        """
        if not self.uuid:
            raise NoIdError("No ID/UUID provided for Extend")
        if not minutes or not isinstance(minutes, int):
            raise IncorrectMinutesValue("Minutes should be an integer")
        await self.check_controller_auth()

        api, method = self.controller.api_extend(self.uuid, minutes=minutes)
        logger.debug(f"API: {method} to {api}")
        headers: Dict[str, str] = {"Accept": "application/json"}
        self.controller.with_auth(headers)

        async with AsyncClient() as client:
            client.headers.update({"User-Agent": self.user_agent})
            r = await client.request(method, api, headers=headers)
            logger.debug(f"Got response: {r.status_code} {r.reason_phrase}")
            if r.status_code < 200 or r.status_code > 399:
                r.raise_for_status()
            return

    @check_controller
    async def wait_for_status(
        self,
        wanted_state: Union[List[IDACRequestStatus], List[str],
                            List[Union[IDACRequestStatus, str]]] = REQUEST_GOOD_STATES,
        stop_on_error: bool = True,
        timeout: Optional[int] = DEFAULT_WAIT_TIMEOUT,
        interval: int = DEFAULT_WAIT_INTERVAL,
    ) -> Union[RequestState, None]:
        """Waits for request to land in one of wanted states by periodically checking it's status

        Args:
            wanted_state (List[Union[IDACRequestStatus, str]], optional): List of wanted statuses.
                Defaults to [ IDACRequestStatus.active, IDACRequestStatus.executed,
                IDACRequestStatus.complete, ].
            stop_on_error (bool, optional): Stop waiting if request errored. Defaults to True.
            timeout (int, optional): Timeout (in seconds). Max amout of time process should wait.
                If set to None will wait forever. Defaults to 10 minutes.
            interval (int, optional): Interval (seconds) between request. Defaults to 30 seconds.

        Raises:
            NoIdError: if no ID/UUID set
            IDACRequestStateError: if request is in errored status
            IncorrectWantedStateError: if incorrect wanted state provided
        """
        if not self.uuid:
            raise NoIdError("No ID/UUID provided for Wait For Status")

        wanted_state = self.sanitize_states(wanted_state)

        async def try_infinite() -> RequestState:
            while True:
                try:
                    st = await self.get_state()
                    status = IDACRequestStatus(st.status)
                    if status in wanted_state:
                        return st
                    if stop_on_error and status in REQUEST_ERROR_STATES:
                        raise IDACRequestStateError(f"Request landed in error state: {status.value}")
                except ReadTimeout as e:
                    pass
                await asyncio.sleep(interval)

        return await asyncio.wait_for(try_infinite(), timeout=timeout)



from __future__ import annotations
from typing import Any, List, Literal, Optional, Union, cast
from pydantic import BaseModel, Extra

from idac_sdk.models.new_request import NewRequest


class LoggerDetails(BaseModel, extra=Extra.allow):
    apiLogToken: Optional[str]
    method: Optional[str]
    api: Optional[str]
    requestId: Optional[str]
    owner: Optional[str]
    recipePath: Optional[str]
    recipeName: Optional[str]


class Vpn(BaseModel):
    username: Optional[str]
    host: Optional[str]
    endpoints: Optional[List[Any]]
    mode: Optional[str]


class Request(NewRequest, extra=Extra.allow):
    api: Optional[str]
    dcloudTimestamp: Optional[str]
    pod: Optional[str]
    uuid: Optional[str]
    cookieLifeHours: Optional[int]
    remote_addr: Optional[str]
    Referer: Optional[str]
    method: Optional[str]
    controller: Optional[str]
    vpn: Optional[Vpn]
    controllerEnv: Optional[str]
    controllerApiUrl: Optional[str]
    user_agent: Optional[str]
    query_string: Optional[str]


class Controller(BaseModel, extra=Extra.allow):
    region: Optional[str]
    webUrl: Optional[str]
    environment: Optional[str]
    extPort: Optional[str]
    location: Optional[str]
    apiUrl: Optional[str]
    cloud: Optional[str]


class Details(BaseModel, extra=Extra.allow):
    language: Optional[str]
    demo: Optional[str]
    name: Optional[str]
    cookieLifeHours: Optional[int]
    notifications: Optional[Any]


class Output(BaseModel, extra=Extra.allow):
    destUrl: Optional[Any]
    loaderPage: Optional[str]
    htmlTemplate: Optional[str]


class Tasks(BaseModel):
    End: Optional[List[Any]]
    Stop: Optional[List[Any]]
    Start: Optional[List[Any]]
    Details: Optional[Details]
    Input: Optional[List[Any]]
    Output: Optional[Output]


class WorkerCode(BaseModel):
    Date: Optional[str]
    Rev: Optional[str]


class Code(BaseModel):
    worker: Optional[WorkerCode]


class Start(BaseModel):
    code: Optional[Code]
    hostname: Optional[str]
    location: Optional[str]


class Worker(BaseModel):
    Start: Optional[Start]


class HistoryItem(BaseModel):
    demo: Optional[str]
    status: Optional[str]
    comments: Optional[str]
    time: Optional[str]


class RequestState(BaseModel, extra=Extra.allow):
    status: Optional[str]
    forcecleanupApi: Optional[str]
    mLoaderUrl: Optional[str]
    loggerDetails: Optional[LoggerDetails]
    workerId: Optional[str]
    registerAction: Optional[str]
    request: Optional[Request]
    loaderUrl: Optional[str]
    comments: Optional[str]
    statusUrl: Optional[str]
    controller: Optional[Controller]
    tasks: Optional[Tasks]
    outputUrl: Optional[str]
    errorUrl: Optional[str]
    cleanupApi: Optional[str]
    restartApi: Optional[str]
    action: Optional[str]
    output: Optional[str]
    vpn: Optional[bool]
    statusApi: Optional[str]
    worker: Optional[Worker]
    history: Optional[List[HistoryItem]]

    def get_task_outputs(
        self, section: Literal["Start", "Stop"], task_name: str
    ) -> Union[dict[str, Any], None]:
        if not hasattr(self.tasks, section):
            raise KeyError(f"No section {section} in tasks")

        for tsk in cast(list, getattr(self.tasks, section)):
            if tsk["name"] == task_name:
                return tsk["output"]

        return None



import asyncclick as click
import json
from jsonpath_ng import parse

from idac_sdk import (
    IDACRequestAsync,
    IDACRequestType,
    SessionData,
)
from idac_sdk.log import logger
from idac_sdk.cli.lib._controller_options import _controller_options, with_controller
from idac_sdk.cli.lib._vpn_options import _vpn_options, with_vpn
from idac_sdk.cli.lib._helpers import (
    add_options,
    filter_dict,
    parse_data,
    parse_json,
    default_output,
    _debug_options,
    set_logger_level,
    append_creds,
)


@click.group()
def commands():
    pass


@commands.command(short_help="execute a recipe")
@add_options(_controller_options)
@add_options(_debug_options)
@click.option(
    "--no-session-xml",
    is_flag=True,
    default=False,
    help="Won't try load session.xml file",
)
@click.option(
    "-s",
    "--session-xml-path",
    default=None,
    type=str,
    help="Path to session.xml file (if non-default should be used)",
)
@click.option(
    "-p",
    "--param",
    help="Additional param to be sent, multiple allowed. Format: KEY=VALUE",
    multiple=True,
    callback=parse_data,
)
@click.option(
    "--skip",
    help="Skip parameter from session.xml, multiple allowed. Format: KEY",
    multiple=True,
)
@click.option(
    "-j",
    "--json",
    help="JSON object to be sent",
    callback=parse_json,
)
@click.option(
    "-t",
    "--type",
    type=click.Choice([e.name for e in IDACRequestType], case_sensitive=False),
    help="Type of the request",
)
@click.option("-o", "--owner", help="Owner", type=str, default=None)
@click.option("-n", "--name", help="Request name (demo name)", type=str, default=None)
@click.option("-d", "--datacenter", help="Datacenter", type=str, default=None)
@click.option(
    "--out-path",
    multiple=True,
    help="Extract value from response by JSON path, Multiple allowed",
    type=str,
    default=None,
)
@click.option(
    "-O",
    "--output",
    help="Would write result to a file.",
    is_flag=True,
    default=False,
)
@click.option(
    "--output-file",
    help=
    r"Sets output file. Default ('C:\dcloud\idac_sdk.out' or '/dcloud/idac_sdk.out') would be used if omitted.",
    type=str,
    default=None,
    callback=default_output)
@click.option(
    "--creds/--no-creds",
    default=True,
    help="Should add dCloud creds to received URL or not. Default: TRUE",
)
@click.option("-k", "--kitchen", help="iDAC Kitchen sandbox location", type=str, default=None)
@add_options(_vpn_options)
@click.argument("recipe_path")
@click.argument("recipe_name")
@with_controller
@with_vpn
async def create(controller, recipe_path, recipe_name, out_path, vpn, creds, **kwargs):
    """Create new automation request (execute a recipe on iDAC)

    \b
    RECIPE_PATH - path to the recipe
    RECIPE_NAME - name of the recipe
    """
    set_logger_level(kwargs.get("debug", False))

    # remove empty params
    filtered = filter_dict(kwargs, lambda el: el[1] is not None)

    # create dict with initial data
    initial = None
    if "json" in filtered or "param" in filtered:
        initial = dict()
        initial.update(filtered.get("param", {}))
        initial.update(filtered.get("json", {}))

    sd_kwargs = {
        "session_xml_path":
            filtered.get("session_xml_path", None) if not filtered["no_session_xml"] else False,
        "recipePath":
            recipe_path,
        "recipeName":
            recipe_name,
    }

    if initial:
        sd_kwargs["initial_data"] = initial

    # create SessionData
    sd = SessionData(**sd_kwargs)

    if "skip" in filtered and (isinstance(filtered["skip"],
                                          (list, tuple)) or isinstance(filtered["skip"], str)):
        skip = filtered["skip"] if isinstance(filtered["skip"],
                                              (list, tuple)) else [filtered["skip"]]
        logger.debug(f"Skipping fields: {skip}")
        for sk in skip:
            if sd.has(sk):
                sd.delete(sk)

    if "owner" in filtered:
        sd.set("owner", filtered["owner"])

    if "name" in filtered:
        sd.set("demo", filtered["name"])

    if "datacenter" in filtered:
        sd.set("datacenter", filtered["datacenter"])

    if "kitchen" in filtered:
        # user would like to execute recipe in a Kitchen sandbox
        sd.set("location", filtered["kitchen"])
        sd.set("recipeMode", "worker")

    req = IDACRequestAsync(session_data=sd, controller=controller, vpn=vpn)
    state, redirect = await req.create(request_type=IDACRequestType[filtered.get("type", "SIMPLE")])

    if creds and sd.has("creds"):
        state = append_creds(state=state, data=sd)

    state_dict = state.dict()
    if redirect:
        state_dict["original_redirect"] = redirect

    file = "-"
    if filtered.get("output", False):
        file = filtered["output_file"]

    with click.open_file(file, "w") as f:
        if out_path:
            for o in out_path:
                jsonpath_expr = parse(o)
                for match in jsonpath_expr.find(state_dict):
                    f.write(f"{match.value}\n")
        else:
            f.write(json.dumps(state_dict))



import logging
import sys

_logger_name = "idac-sdk"


def create_logger() -> logging.Logger:
    # configure log formatter
    logFormatter = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")

    # configure stream handler
    consoleHandler = logging.StreamHandler(sys.stdout)
    consoleHandler.setFormatter(logFormatter)

    # get the logger instance
    lgr = logging.getLogger(_logger_name)

    # set default logging level
    lgr.setLevel(logging.WARNING)

    # add console handler
    lgr.addHandler(consoleHandler)
    return lgr


logger: logging.Logger = (
    create_logger()
    if not logging.getLogger(_logger_name).hasHandlers()
    else logging.getLogger(_logger_name)
)
